@@ -37,21 +37,21 @@
 #include <sys/types.h>
 #include <sys/xattr.h>
 
-#include "ctree.h"
-#include "disk-io.h"
-#include "print-tree.h"
-#include "transaction.h"
+#include "kernel-shared/ctree.h"
+#include "kernel-shared/disk-io.h"
+#include "kernel-shared/print-tree.h"
+#include "kernel-shared/transaction.h"
 #include "kernel-lib/list.h"
-#include "volumes.h"
+#include "kernel-shared/volumes.h"
 #include "common/utils.h"
 #include "cmds/commands.h"
 #include "common/help.h"
+#include "common/open-utils.h"
 
 static char fs_name[PATH_MAX];
 static char path_name[PATH_MAX];
 static char symlink_target[PATH_MAX];
 static int get_snaps = 0;
-static int verbose = 0;
 static int restore_metadata = 0;
 static int restore_symlinks = 0;
 static int ignore_errors = 0;
@@ -310,8 +310,8 @@ static int copy_one_inline(struct btrfs_root *root, int fd,
 	if (compress == BTRFS_COMPRESS_NONE) {
 		done = pwrite(fd, buf, len, pos);
 		if (done < len) {
-			fprintf(stderr, "Short inline write, wanted %d, did "
-				"%zd: %d\n", len, done, errno);
+			error("short inline write, wanted %d, did %zd: %m",
+					len, done);
 			return -1;
 		}
 		return 0;
@@ -334,8 +334,8 @@ static int copy_one_inline(struct btrfs_root *root, int fd,
 	done = pwrite(fd, outbuf, ram_size, pos);
 	free(outbuf);
 	if (done < ram_size) {
-		fprintf(stderr, "Short compressed inline write, wanted %Lu, "
-			"did %zd: %d\n", ram_size, done, errno);
+		error("short compressed inline write, wanted %llu, did %zd: %m",
+				ram_size, done);
 		return -1;
 	}
 
@@ -346,8 +346,6 @@ static int copy_one_extent(struct btrfs_root *root, int fd,
 			   struct extent_buffer *leaf,
 			   struct btrfs_file_extent_item *fi, u64 pos)
 {
-	struct btrfs_multi_bio *multi = NULL;
-	struct btrfs_device *device;
 	char *inbuf, *outbuf = NULL;
 	ssize_t done, total = 0;
 	u64 bytenr;
@@ -356,12 +354,10 @@ static int copy_one_extent(struct btrfs_root *root, int fd,
 	u64 num_bytes;
 	u64 length;
 	u64 size_left;
-	u64 dev_bytenr;
 	u64 offset;
-	u64 count = 0;
+	u64 cur;
 	int compress;
 	int ret;
-	int dev_fd;
 	int mirror_num = 1;
 	int num_copies;
 
@@ -372,15 +368,26 @@ static int copy_one_extent(struct btrfs_root *root, int fd,
 	offset = btrfs_file_extent_offset(leaf, fi);
 	num_bytes = btrfs_file_extent_num_bytes(leaf, fi);
 	size_left = disk_size;
-	if (compress == BTRFS_COMPRESS_NONE)
-		bytenr += offset;
-
-	if (verbose && offset)
-		printf("offset is %Lu\n", offset);
-	/* we found a hole */
+	/* Hole, early exit */
 	if (disk_size == 0)
 		return 0;
 
+	/* Invalid file extent */
+	if ((compress == BTRFS_COMPRESS_NONE && offset >= disk_size) ||
+	    offset > ram_size) {
+		error(
+	"invalid data extent offset, offset %llu disk_size %llu ram_size %llu",
+		      offset, disk_size, ram_size);
+		return -EUCLEAN;
+	}
+
+	if (compress == BTRFS_COMPRESS_NONE && offset < disk_size) {
+		bytenr += offset;
+		size_left -= offset;
+	}
+
+	pr_verbose(offset ? 1 : 0, "offset is %llu\n", offset);
+
 	inbuf = malloc(size_left);
 	if (!inbuf) {
 		error("not enough memory");
@@ -395,48 +402,28 @@ static int copy_one_extent(struct btrfs_root *root, int fd,
 			return -ENOMEM;
 		}
 	}
+
+	num_copies = btrfs_num_copies(root->fs_info, bytenr, disk_size - offset);
 again:
-	length = size_left;
-	ret = btrfs_map_block(root->fs_info, READ, bytenr, &length, &multi,
-			      mirror_num, NULL);
-	if (ret) {
-		error("cannot map block logical %llu length %llu: %d",
-				(unsigned long long)bytenr,
-				(unsigned long long)length, ret);
-		goto out;
-	}
-	device = multi->stripes[0].dev;
-	dev_fd = device->fd;
-	device->total_ios++;
-	dev_bytenr = multi->stripes[0].physical;
-	free(multi);
-
-	if (size_left < length)
-		length = size_left;
-
-	done = pread(dev_fd, inbuf+count, length, dev_bytenr);
-	/* Need both checks, or we miss negative values due to u64 conversion */
-	if (done < 0 || done < length) {
-		num_copies = btrfs_num_copies(root->fs_info, bytenr, length);
-		mirror_num++;
-		/* mirror_num is 1-indexed, so num_copies is a valid mirror. */
-		if (mirror_num > num_copies) {
-			ret = -1;
-			error("exhausted mirrors trying to read (%d > %d)",
+	cur = bytenr;
+	while (cur < bytenr + size_left) {
+		length = bytenr + size_left - cur;
+		ret = read_extent_data(root->fs_info, inbuf + cur - bytenr, cur,
+				       &length, mirror_num);
+		if (ret < 0) {
+			mirror_num++;
+			if (mirror_num > num_copies) {
+				ret = -1;
+				error("exhausted mirrors trying to read (%d > %d)",
 					mirror_num, num_copies);
-			goto out;
+				goto out;
+			}
+			fprintf(stderr, "trying another mirror\n");
+			continue;
 		}
-		fprintf(stderr, "Trying another mirror\n");
-		goto again;
+		cur += length;
 	}
 
-	mirror_num = 1;
-	size_left -= length;
-	count += length;
-	bytenr += length;
-	if (size_left)
-		goto again;
-
 	if (compress == BTRFS_COMPRESS_NONE) {
 		while (total < num_bytes) {
 			done = pwrite(fd, inbuf+total, num_bytes-total,
@@ -454,13 +441,13 @@ again:
 
 	ret = decompress(root, inbuf, outbuf, disk_size, &ram_size, compress);
 	if (ret) {
-		num_copies = btrfs_num_copies(root->fs_info, bytenr, length);
 		mirror_num++;
-		if (mirror_num >= num_copies) {
+		if (mirror_num > num_copies) {
 			ret = -1;
 			goto out;
 		}
-		fprintf(stderr, "Trying another mirror\n");
+		fprintf(stderr,
+			"trying another mirror due to decompression error\n");
 		goto again;
 	}
 
@@ -480,34 +467,6 @@ out:
 	return ret;
 }
 
-enum loop_response {
-	LOOP_STOP,
-	LOOP_CONTINUE,
-	LOOP_DONTASK
-};
-
-static enum loop_response ask_to_continue(const char *file)
-{
-	char buf[16];
-	char *ret;
-
-	printf("We seem to be looping a lot on %s, do you want to keep going "
-	       "on ? (y/N/a): ", file);
-again:
-	ret = fgets(buf, 16, stdin);
-	if (!ret || *ret == '\n' || tolower(*ret) == 'n')
-		return LOOP_STOP;
-	if (tolower(*ret) == 'a')
-		return LOOP_DONTASK;
-	if (tolower(*ret) != 'y') {
-		printf("Please enter one of 'y', 'n', or 'a': ");
-		goto again;
-	}
-
-	return LOOP_CONTINUE;
-}
-
-
 static int set_file_xattrs(struct btrfs_root *root, u64 inode,
 			   int fd, const char *file_name)
 {
@@ -666,7 +625,6 @@ static int copy_file(struct btrfs_root *root, int fd, struct btrfs_key *key,
 	int ret;
 	int extent_type;
 	int compression;
-	int loops = 0;
 	u64 found_size = 0;
 	struct timespec times[2];
 	int times_ok = 0;
@@ -729,22 +687,11 @@ static int copy_file(struct btrfs_root *root, int fd, struct btrfs_key *key,
 	}
 
 	while (1) {
-		if (loops >= 0 && loops++ >= 1024) {
-			enum loop_response resp;
-
-			resp = ask_to_continue(file);
-			if (resp == LOOP_STOP)
-				break;
-			else if (resp == LOOP_CONTINUE)
-				loops = 0;
-			else if (resp == LOOP_DONTASK)
-				loops = -1;
-		}
 		if (path.slots[0] >= btrfs_header_nritems(leaf)) {
 			do {
 				ret = next_leaf(root, &path);
 				if (ret < 0) {
-					fprintf(stderr, "Error searching %d\n", ret);
+					error("search to next leaf failed: %d", ret);
 					goto out;
 				} else if (ret) {
 					/* No more leaves to search */
@@ -832,11 +779,13 @@ static int overwrite_ok(const char * path)
 		if (overwrite)
 			return 2;
 
-		if (verbose || !warn)
-			printf("Skipping existing file"
-				   " %s\n", path);
-		if (!warn)
-			printf("If you wish to overwrite use -o\n");
+		if (!warn) {
+			pr_verbose(-1, "Skipping existing file %s\n", path);
+			pr_verbose(-1, "If you wish to overwrite use -o\n");
+		} else {
+			pr_verbose(1, "Skipping existing file %s\n", path);
+		}
+
 		warn = 1;
 		return 0;
 	}
@@ -864,8 +813,7 @@ static int copy_symlink(struct btrfs_root *root, struct btrfs_key *key,
 	if (ret == 2) {
 		ret = unlink(path_name);
 		if (ret) {
-			fprintf(stderr, "failed to unlink '%s' for overwrite\n",
-					path_name);
+			error("failed to unlink '%s' for overwrite: %m", path_name);
 			return ret;
 		}
 	}
@@ -879,7 +827,7 @@ static int copy_symlink(struct btrfs_root *root, struct btrfs_key *key,
 
 	leaf = path.nodes[0];
 	if (!leaf) {
-		fprintf(stderr, "Error getting leaf for symlink '%s'\n", file);
+		error("failed to get leaf for symlink '%s'", file);
 		ret = -1;
 		goto out;
 	}
@@ -890,7 +838,7 @@ static int copy_symlink(struct btrfs_root *root, struct btrfs_key *key,
 	len = btrfs_file_extent_inline_item_len(leaf,
 			btrfs_item_nr(path.slots[0]));
 	if (len >= PATH_MAX) {
-		fprintf(stderr, "Symlink '%s' target length %d is longer than PATH_MAX\n",
+		error("symlink '%s' target length %d is longer than PATH_MAX",
 				fs_name, len);
 		ret = -1;
 		goto out;
@@ -905,12 +853,13 @@ static int copy_symlink(struct btrfs_root *root, struct btrfs_key *key,
 	if (!dry_run) {
 		ret = symlink(symlink_target, path_name);
 		if (ret < 0) {
-			fprintf(stderr, "Failed to restore symlink '%s': %m\n",
-					path_name);
+			error("failed to restore symlink '%s': %m", path_name);
 			goto out;
 		}
 	}
-	printf("SYMLINK: '%s' => '%s'\n", path_name, symlink_target);
+
+	if (bconf.verbose >= 2)
+		printf("SYMLINK: '%s' => '%s'\n", path_name, symlink_target);
 
 	ret = 0;
 	if (!restore_metadata)
@@ -927,18 +876,18 @@ static int copy_symlink(struct btrfs_root *root, struct btrfs_key *key,
 
 	ret = btrfs_lookup_inode(NULL, root, &path, key, 0);
 	if (ret) {
-		fprintf(stderr, "Failed to lookup inode for '%s'\n", file);
+		error("failed to lookup inode for '%s'", file);
 		goto out;
 	}
 
 	inode_item = btrfs_item_ptr(path.nodes[0], path.slots[0],
 			struct btrfs_inode_item);
 
-	ret = fchownat(-1, file, btrfs_inode_uid(path.nodes[0], inode_item),
+	ret = fchownat(AT_FDCWD, file, btrfs_inode_uid(path.nodes[0], inode_item),
 				   btrfs_inode_gid(path.nodes[0], inode_item),
 				   AT_SYMLINK_NOFOLLOW);
 	if (ret) {
-		fprintf(stderr, "Failed to change owner: %m\n");
+		error("failed to change owner of '%s': %m", file);
 		goto out;
 	}
 
@@ -950,9 +899,9 @@ static int copy_symlink(struct btrfs_root *root, struct btrfs_key *key,
 	times[1].tv_sec  = btrfs_timespec_sec(path.nodes[0], bts);
 	times[1].tv_nsec = btrfs_timespec_nsec(path.nodes[0], bts);
 
-	ret = utimensat(-1, file, times, AT_SYMLINK_NOFOLLOW);
+	ret = utimensat(AT_FDCWD, file, times, AT_SYMLINK_NOFOLLOW);
 	if (ret)
-		fprintf(stderr, "Failed to set times: %m\n");
+		error("failed to set times for '%s': %m", file);
 out:
 	btrfs_release_path(&path);
 	return ret;
@@ -971,7 +920,6 @@ static int search_dir(struct btrfs_root *root, struct btrfs_key *key,
 	int name_len;
 	int ret = 0;
 	int fd;
-	int loops = 0;
 	u8 type;
 
 	btrfs_init_path(&path);
@@ -979,7 +927,7 @@ static int search_dir(struct btrfs_root *root, struct btrfs_key *key,
 	key->type = BTRFS_DIR_INDEX_KEY;
 	ret = btrfs_search_slot(NULL, root, key, &path, 0, 0);
 	if (ret < 0) {
-		fprintf(stderr, "Error searching %d\n", ret);
+		error("search for next directory entry failed: %d", ret);
 		goto out;
 	}
 
@@ -987,19 +935,16 @@ static int search_dir(struct btrfs_root *root, struct btrfs_key *key,
 
 	leaf = path.nodes[0];
 	while (!leaf) {
-		if (verbose > 1)
-			printf("No leaf after search, looking for the next "
-			       "leaf\n");
+		pr_verbose(2,
+			   "No leaf after search, looking for the next leaf\n");
 		ret = next_leaf(root, &path);
 		if (ret < 0) {
-			fprintf(stderr, "Error getting next leaf %d\n",
-				ret);
+			error("search for next leaf failed: %d", ret);
 			goto out;
 		} else if (ret > 0) {
 			/* No more leaves to search */
-			if (verbose)
-				printf("Reached the end of the tree looking "
-				       "for the directory\n");
+			pr_verbose(1,
+		   "Reached the end of the tree looking for the directory\n");
 			ret = 0;
 			goto out;
 		}
@@ -1007,26 +952,16 @@ static int search_dir(struct btrfs_root *root, struct btrfs_key *key,
 	}
 
 	while (leaf) {
-		if (loops++ >= 1024) {
-			printf("We have looped trying to restore files in %s "
-			       "too many times to be making progress, "
-			       "stopping\n", in_dir);
-			break;
-		}
-
 		if (path.slots[0] >= btrfs_header_nritems(leaf)) {
 			do {
 				ret = next_leaf(root, &path);
 				if (ret < 0) {
-					fprintf(stderr, "Error searching %d\n",
-						ret);
+					error("search for next leaf faile: %d", ret);
 					goto out;
 				} else if (ret > 0) {
 					/* No more leaves to search */
-					if (verbose)
-						printf("Reached the end of "
-						       "the tree searching the"
-						       " directory\n");
+					pr_verbose(1,
+		"Reached the end of the tree searching the directory\n");
 					ret = 0;
 					goto out;
 				}
@@ -1036,14 +971,12 @@ static int search_dir(struct btrfs_root *root, struct btrfs_key *key,
 		}
 		btrfs_item_key_to_cpu(leaf, &found_key, path.slots[0]);
 		if (found_key.objectid != key->objectid) {
-			if (verbose > 1)
-				printf("Found objectid=%Lu, key=%Lu\n",
-				       found_key.objectid, key->objectid);
+			pr_verbose(2, "Found objectid=%llu, key=%llu\n",
+				   found_key.objectid, key->objectid);
 			break;
 		}
 		if (found_key.type != key->type) {
-			if (verbose > 1)
-				printf("Found type=%u, want=%u\n",
+			pr_verbose(2, "Found type=%u, want=%u\n",
 				       found_key.type, key->type);
 			break;
 		}
@@ -1072,25 +1005,21 @@ static int search_dir(struct btrfs_root *root, struct btrfs_key *key,
 			if (!overwrite_ok(path_name))
 				goto next;
 
-			if (verbose)
-				printf("Restoring %s\n", path_name);
+			pr_verbose(1, "Restoring %s\n", path_name);
 			if (dry_run)
 				goto next;
 			fd = open(path_name, O_CREAT|O_WRONLY, 0644);
 			if (fd < 0) {
-				fprintf(stderr, "Error creating %s: %d\n",
-					path_name, errno);
+				error("creating '%s' failed: %m", path_name);
 				if (ignore_errors)
 					goto next;
 				ret = -1;
 				goto out;
 			}
-			loops = 0;
 			ret = copy_file(root, fd, &location, path_name);
 			close(fd);
 			if (ret) {
-				fprintf(stderr, "Error copying data for %s\n",
-					path_name);
+				error("copying data for %s failed", path_name);
 				if (ignore_errors)
 					goto next;
 				goto out;
@@ -1100,7 +1029,7 @@ static int search_dir(struct btrfs_root *root, struct btrfs_key *key,
 			char *dir = strdup(fs_name);
 
 			if (!dir) {
-				fprintf(stderr, "Ran out of memory\n");
+				error("ran out of memory");
 				ret = -ENOMEM;
 				goto out;
 			}
@@ -1121,10 +1050,8 @@ static int search_dir(struct btrfs_root *root, struct btrfs_key *key,
 								 &location);
 				if (IS_ERR(search_root)) {
 					free(dir);
-					fprintf(stderr, "Error reading "
-						"subvolume %s: %lu\n",
-						path_name,
-						PTR_ERR(search_root));
+					error("reading subvolume %s failed: %lu",
+						path_name, PTR_ERR(search_root));
 					if (ignore_errors)
 						goto next;
 					ret = PTR_ERR(search_root);
@@ -1138,15 +1065,13 @@ static int search_dir(struct btrfs_root *root, struct btrfs_key *key,
 				if (search_root->root_key.offset != 0 &&
 				    get_snaps == 0) {
 					free(dir);
-					printf("Skipping snapshot %s\n",
-					       filename);
+					printf("Skipping snapshot %s\n", filename);
 					goto next;
 				}
 				location.objectid = BTRFS_FIRST_FREE_OBJECTID;
 			}
 
-			if (verbose)
-				printf("Restoring %s\n", path_name);
+			pr_verbose(1, "Restoring %s\n", path_name);
 
 			errno = 0;
 			if (dry_run)
@@ -1155,20 +1080,18 @@ static int search_dir(struct btrfs_root *root, struct btrfs_key *key,
 				ret = mkdir(path_name, 0755);
 			if (ret && errno != EEXIST) {
 				free(dir);
-				fprintf(stderr, "Error mkdiring %s: %d\n",
-					path_name, errno);
+				error("failed mkdir %s: %m", path_name);
 				if (ignore_errors)
 					goto next;
 				ret = -1;
 				goto out;
 			}
-			loops = 0;
 			ret = search_dir(search_root, &location,
 					 output_rootdir, dir, mreg);
 			free(dir);
 			if (ret) {
-				fprintf(stderr, "Error searching %s\n",
-					path_name);
+				error("searching directory %s failed: %d",
+					path_name, ret);
 				if (ignore_errors)
 					goto next;
 				goto out;
@@ -1191,7 +1114,7 @@ next:
 		snprintf(path_name, PATH_MAX, "%s%s", output_rootdir, in_dir);
 		fd = open(path_name, O_RDONLY);
 		if (fd < 0) {
-			fprintf(stderr, "ERROR: Failed to access %s to restore metadata\n",
+			error("failed to access '%s' to restore metadata: %m",
 					path_name);
 			if (!ignore_errors) {
 				ret = -1;
@@ -1209,8 +1132,7 @@ next:
 		}
 	}
 
-	if (verbose)
-		printf("Done searching %s\n", in_dir);
+	pr_verbose(1, "Done searching %s\n", in_dir);
 out:
 	btrfs_release_path(&path);
 	return ret;
@@ -1236,7 +1158,7 @@ static int do_list_roots(struct btrfs_root *root)
 	key.type = BTRFS_ROOT_ITEM_KEY;
 	ret = btrfs_search_slot(NULL, root, &key, &path, 0, 0);
 	if (ret < 0) {
-		fprintf(stderr, "Failed to do search %d\n", ret);
+		error("failed search next root item: %d", ret);
 		btrfs_release_path(&path);
 		return -1;
 	}
@@ -1263,7 +1185,7 @@ static int do_list_roots(struct btrfs_root *root)
 		read_extent_buffer(leaf, &ri, offset, sizeof(ri));
 		printf(" tree ");
 		btrfs_print_key(&disk_key);
-		printf(" %Lu level %d\n", btrfs_root_bytenr(&ri),
+		printf(" %llu level %d\n", btrfs_root_bytenr(&ri),
 		       btrfs_root_level(&ri));
 		path.slots[0]++;
 	}
@@ -1277,6 +1199,7 @@ static struct btrfs_root *open_fs(const char *dev, u64 root_location,
 {
 	struct btrfs_fs_info *fs_info = NULL;
 	struct btrfs_root *root = NULL;
+	struct open_ctree_flags ocf = { 0 };
 	u64 bytenr;
 	int i;
 
@@ -1288,9 +1211,11 @@ static struct btrfs_root *open_fs(const char *dev, u64 root_location,
 		 * in extent tree. Skip block group item search will allow
 		 * restore to be executed on heavily damaged fs.
 		 */
-		fs_info = open_ctree_fs_info(dev, bytenr, root_location, 0,
-					     OPEN_CTREE_PARTIAL |
-					     OPEN_CTREE_NO_BLOCK_GROUPS);
+		ocf.filename = dev;
+		ocf.sb_bytenr = bytenr;
+		ocf.root_tree_bytenr = root_location;
+		ocf.flags = OPEN_CTREE_PARTIAL | OPEN_CTREE_NO_BLOCK_GROUPS;
+		fs_info = open_ctree_fs_info(&ocf);
 		if (fs_info)
 			break;
 		fprintf(stderr, "Could not open root, trying backup super\n");
@@ -1314,7 +1239,7 @@ static struct btrfs_root *open_fs(const char *dev, u64 root_location,
 		root->node = read_tree_block(fs_info, root_location,
 					     generation);
 		if (!extent_buffer_uptodate(root->node)) {
-			fprintf(stderr, "Error opening tree root\n");
+			error("opening tree root failed");
 			close_ctree(root);
 			return NULL;
 		}
@@ -1328,8 +1253,7 @@ static struct btrfs_root *open_fs(const char *dev, u64 root_location,
 		key.offset = (u64)-1;
 		fs_info->fs_root = btrfs_read_fs_root_no_cache(fs_info, &key);
 		if (IS_ERR(fs_info->fs_root)) {
-			fprintf(stderr, "Couldn't read fs root: %ld\n",
-				PTR_ERR(fs_info->fs_root));
+			error("could not read fs root: %ld", PTR_ERR(fs_info->fs_root));
 			close_ctree(fs_info->tree_root);
 			return NULL;
 		}
@@ -1357,12 +1281,12 @@ static int find_first_dir(struct btrfs_root *root, u64 *objectid)
 	key.offset = 0;
 	ret = btrfs_search_slot(NULL, root, &key, &path, 0, 0);
 	if (ret < 0) {
-		fprintf(stderr, "Error searching %d\n", ret);
+		error("searching next directory entry failed: %d", ret);
 		goto out;
 	}
 
 	if (!path.nodes[0]) {
-		fprintf(stderr, "No leaf!\n");
+		error("no leaf when looking for directory");
 		goto out;
 	}
 again:
@@ -1372,7 +1296,7 @@ again:
 		if (found_key.type != key.type)
 			continue;
 
-		printf("Using objectid %Lu for first dir\n",
+		printf("Using objectid %llu for first dir\n",
 		       found_key.objectid);
 		*objectid = found_key.objectid;
 		ret = 0;
@@ -1381,11 +1305,10 @@ again:
 	do {
 		ret = next_leaf(root, &path);
 		if (ret < 0) {
-			fprintf(stderr, "Error getting next leaf %d\n",
-				ret);
+			error("search for next leaf failed: %d", ret);
 			goto out;
 		} else if (ret > 0) {
-			fprintf(stderr, "No more leaves\n");
+			error("no more leaves to search");
 			goto out;
 		}
 	} while (!path.nodes[0]);
@@ -1398,28 +1321,37 @@ out:
 }
 
 static const char * const cmd_restore_usage[] = {
-	"btrfs restore [options] <device> <path> | -l <device>",
+	"btrfs restore [options] <device> <path>\n"
+	"btrfs restore [options] -l <device>",
 	"Try to restore files from a damaged filesystem (unmounted)",
 	"",
-	"-s|--snapshots       get snapshots",
-	"-x|--xattr           restore extended attributes",
-	"-m|--metadata        restore owner, mode and times",
-	"-S|--symlink         restore symbolic links",
-	"-v|--verbose         verbose",
-	"-i|--ignore-errors   ignore errors",
-	"-o|--overwrite       overwrite",
-	"-t <bytenr>          tree location",
-	"-f <bytenr>          filesystem location",
-	"-u|--super <mirror>  super mirror",
-	"-r|--root <rootid>   root objectid",
-	"-d                   find dir",
-	"-l|--list-roots      list tree roots",
-	"-D|--dry-run         dry run (only list files that would be recovered)",
-	"--path-regex <regex>",
-	"                     restore only filenames matching regex,",
-	"                     you have to use following syntax (possibly quoted):",
-	"                     ^/(|home(|/username(|/Desktop(|/.*))))$",
-	"-c                   ignore case (--path-regex only)",
+	"  control:",
+	"       -D|--dry-run         dry run (only list files that would be recovered)",
+	"       -i|--ignore-errors   ignore errors",
+	"       -o|--overwrite       overwrite",
+	"  restoration:",
+	"       -m|--metadata        restore owner, mode and times",
+	"       -S|--symlink         restore symbolic links",
+	"       -s|--snapshots       get snapshots",
+	"       -x|--xattr           restore extended attributes",
+	"  filtering:",
+	"       --path-regex <regex>",
+	"                            restore only filenames matching regex,",
+	"                            you have to use following syntax (possibly quoted):",
+	"                            ^/(|home(|/username(|/Desktop(|/.*))))$",
+	"       -c                   ignore case (--path-regex only)",
+	"  analysis:",
+	"       -d                   find dir",
+	"       -l|--list-roots      list tree roots",
+	"  alternate starting point:",
+	"       -f <bytenr>          filesystem location",
+	"       -r|--root <rootid>   root objectid",
+	"       -t <bytenr>          tree location",
+	"       -u|--super <mirror>  super mirror",
+	"  other:",
+	"       -v|--verbose         deprecated, alias for global -v option",
+	HELPINFO_INSERT_GLOBALS,
+	HELPINFO_INSERT_VERBOSE,
 	NULL
 };
 
@@ -1472,7 +1404,7 @@ static int cmd_restore(const struct cmd_struct *cmd, int argc, char **argv)
 				get_snaps = 1;
 				break;
 			case 'v':
-				verbose++;
+				bconf_be_verbose();
 				break;
 			case 'i':
 				ignore_errors = 1;
@@ -1489,8 +1421,8 @@ static int cmd_restore(const struct cmd_struct *cmd, int argc, char **argv)
 			case 'u':
 				super_mirror = arg_strtou64(optarg);
 				if (super_mirror >= BTRFS_SUPER_MIRROR_MAX) {
-					fprintf(stderr, "Super mirror not "
-						"valid\n");
+					error("super mirror %d not valid",
+							super_mirror);
 					exit(1);
 				}
 				break;
@@ -1500,7 +1432,7 @@ static int cmd_restore(const struct cmd_struct *cmd, int argc, char **argv)
 			case 'r':
 				root_objectid = arg_strtou64(optarg);
 				if (!is_fstree(root_objectid)) {
-					fprintf(stderr, "objectid %llu is not a valid fs/file tree\n",
+					error("objectid %llu is not a valid fs/file tree",
 							root_objectid);
 					exit(1);
 				}
@@ -1537,16 +1469,16 @@ static int cmd_restore(const struct cmd_struct *cmd, int argc, char **argv)
 		usage(cmd);
 
 	if (fs_location && root_objectid) {
-		fprintf(stderr, "don't use -f and -r at the same time.\n");
+		error("can't use -f and -r at the same time");
 		return 1;
 	}
 
 	if ((ret = check_mounted(argv[optind])) < 0) {
 		errno = -ret;
-		fprintf(stderr, "Could not check mount status: %m\n");
+		error("could not check mount status: %m");
 		return 1;
 	} else if (ret) {
-		fprintf(stderr, "%s is currently mounted.  Aborting.\n", argv[optind]);
+		error("%s is currently mounted, cannot continue", argv[optind]);
 		return 1;
 	}
 
@@ -1561,7 +1493,7 @@ static int cmd_restore(const struct cmd_struct *cmd, int argc, char **argv)
 		free_extent_buffer(root->node);
 		root->node = read_tree_block(root->fs_info, fs_location, 0);
 		if (!extent_buffer_uptodate(root->node)) {
-			fprintf(stderr, "Failed to read fs location\n");
+			error("failed to read fs location");
 			ret = 1;
 			goto out;
 		}
@@ -1570,7 +1502,7 @@ static int cmd_restore(const struct cmd_struct *cmd, int argc, char **argv)
 	memset(path_name, 0, PATH_MAX);
 
 	if (strlen(argv[optind + 1]) >= PATH_MAX) {
-		fprintf(stderr, "ERROR: path too long\n");
+		error("path '%s' too long", argv[optind + 1]);
 		ret = 1;
 		goto out;
 	}
@@ -1592,8 +1524,7 @@ static int cmd_restore(const struct cmd_struct *cmd, int argc, char **argv)
 		root = btrfs_read_fs_root(orig_root->fs_info, &key);
 		if (IS_ERR(root)) {
 			errno = -PTR_ERR(root);
-			fprintf(stderr, "fail to read root %llu: %m\n",
-					root_objectid);
+			error("failed to read root %llu: %m", root_objectid);
 			root = orig_root;
 			ret = 1;
 			goto out;
@@ -1614,7 +1545,7 @@ static int cmd_restore(const struct cmd_struct *cmd, int argc, char **argv)
 		ret = regcomp(&match_reg, match_regstr, match_cflags);
 		if (ret) {
 			regerror(ret, &match_reg, reg_err, sizeof(reg_err));
-			fprintf(stderr, "Regex compile failed: %s\n", reg_err);
+			error("regex compilation failed: %s", reg_err);
 			goto out;
 		}
 		mreg = &match_reg;
